#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

import plistlib
import subprocess
import sys
import time

SLEEP_DURATION_SECS = 20
TEST_DURATION_SECS = 20

class DisplayInfo:
    """
    Holds information about a specific display interface.
    """

    def __init__(self, is_external, is_connected):
        self.is_external = is_external
        self.is_connected = is_connected
        self.product = None
        self.transport = None
        self.width = None
        self.height = None
        self.disconnect_start = None
        self.disconnect_length = None

    @classmethod
    def get_current(cls):
        """
        Returns information on the display interfaces, as a dictionary.
        """

        displays = {}

        for props in cls._get_ioreg_info():
            ident = props['IONameMatched']

            # Attribute (currently?) not present on internal displays
            is_external = props.get('external', False)

            # Attribute present only on connected external displays, and
            # “active” internal displays (i.e. lid not closed)
            is_connected = 'TimingElements' in props

            info = cls(is_external, is_connected)

            attrs = props.get('DisplayAttributes')

            info.product = cls._get_opt(attrs, 'ProductAttributes',
                'ProductName')
            info.transport = cls._get_opt(props, 'Transport', 'Downstream')
            info.width = cls._get_opt(attrs, 'NativeFormatHorizontalPixels')
            info.height = cls._get_opt(attrs, 'NativeFormatVerticalPixels')

            displays[ident] = info

        return displays

    @classmethod
    def check_disconnection(cls, displays, start_time, duration):
        """
        Watches the status if the displays every 100ms, for duration seconds.
        Uses the start_time monotonic time as a reference for when the displays
        started to wake up. Fills the disconnect_start and disconnect_length
        attributes of the provided DisplayInfo objects.
        """

        disco_start = {}
        disco_end = {}

        for ident in displays.keys():
            disco_start[ident] = None
            disco_end[ident] = None

        end_time = start_time + duration
        cur_time = time.monotonic()

        while cur_time < end_time:
            new_infos = cls.get_current()

            for ident in displays.keys():
                if not new_infos[ident].is_connected:
                    if disco_start[ident] is None:
                        disco_start[ident] = cur_time

                    disco_end[ident] = cur_time

            cur_time = time.monotonic()

        for ident, display_info in displays.items():
            disp_disco_start = disco_start[ident]
            disp_disco_end = disco_end[ident]
            if disp_disco_start is None:
                display_info.disconnect_start = 0
                display_info.disconnect_length = 0
            else:
                length = disp_disco_end - disp_disco_start
                if length == 0:
                    length = .1

                display_info.disconnect_start = disp_disco_start - start_time
                display_info.disconnect_length = length

    def __str__(self):
        category = "external" if self.is_external else "internal"
        conn = "connected" if self.is_connected else "disconnected"

        if self.product:
            category = f"{category} ({self.product})"

        res = f"{category}, {conn}"
        if self.transport:
            res += f", {self.transport} transport"

        if self.width and self.height:
            res += f", {self.width}x{self.height}"

        if (self.is_external and self.is_connected and
            self.disconnect_start is not None):
            if self.disconnect_length:
                start = int(self.disconnect_start * 1000)
                length = int(self.disconnect_length * 1000)
                res += f", disconnected after {start}ms for {length}ms"
            else:
                res += f", did not disconnect"

        return res

    def __repr__(self):
        return (f"<DisplayInfo ext {self.is_external} conn "
            f"{self.is_connected} product {self.product} transport "
            f"{self.transport} {self.width}x{self.height} disco "
            f"{self.disconnect_start} {self.disconnect_length}")

    @staticmethod
    def _get_opt(props, *path):
        """
        Get an optional property from a dictionary
        """

        cur = props
        for component in path:
            if cur is None:
                return None

            cur = cur.get(component)

        return cur

    @staticmethod
    def _get_ioreg_info():
        """
        Return IOReg AppleCLC2 information, as a Python array of properties.
        """

        proc = subprocess.Popen(['ioreg', '-ra', '-c', 'AppleCLCD2'],
            stdin=subprocess.DEVNULL, stdout=subprocess.PIPE)
        try:
            return plistlib.load(proc.stdout, fmt=plistlib.FMT_XML)
        finally:
            try:
                res = proc.wait(timeout=1)
            except subprocess.TimeoutExpired:
                proc.kill()
                raise


def sleep_displays(duration):
    """
    Sleep the displays for the specified time. Returns the approximative
    monotonic time at which the display woke up.
    """

    subprocess.run(['pmset', 'displaysleepnow'], check=True)
    print("Displays should be sleeping now. If they are not, something is "
        "wrong.", end='', flush=True)
    time.sleep(duration)
    print("\r\x1b[K", end='', flush=True)

    # Pass a command line to caffeinate so it exits immediately.
    # /bin/true is not in the recovery environment so it can’t be used.
    subprocess.run(['caffeinate', '-u',  '/bin/echo', '-n'], check=True)
    return time.monotonic()


def perform_test(mac_model, sleep_duration):
    displays = DisplayInfo.get_current()

    wake_time = sleep_displays(sleep_duration)

    print("Detection in progress, please wait…")
    DisplayInfo.check_disconnection(displays, wake_time, TEST_DURATION_SECS)

    print(f"Test results:")
    print(f"Mac model: {mac_model}")
    print(f"Sleep duration: {sleep_duration}")
    for ident in sorted(displays.keys()):
        print(f"{ident}: {displays[ident]}")

def run():
    mac_model = subprocess.run(['sysctl', '-b', 'hw.model'], check=True,
        stdout=subprocess.PIPE, encoding='utf-8').stdout.rstrip('\x00')

    print("This program will check if your monitors disconnect when getting "
        "out of sleep\nmode.")
    print(" * Your monitors will go blank for the duration specified below.")
    print(" * Do not press a key or move the cursor until the monitors come "
        "back.")
    print(" * If music is playing through one of your external monitors, "
        "pause it before\n   continuing (really).")
    print(" * You may need to unlock your session afterwards.")
    print("")

    durations = [float(value) for value in sys.argv[1:]]

    while not durations:
        try:
            prompt = f"Monitor sleep duration: [{SLEEP_DURATION_SECS}] "
            duration = input(prompt)
            if duration:
                duration = float(duration)
            else:
                duration = SLEEP_DURATION_SECS

        except ValueError:
            duration = 0
        if duration > 0:
            durations = [duration]

    for duration in durations:
        perform_test(mac_model, duration)


if __name__ == '__main__':
    run()
